/* Copyright (C) 2008 by Attila Tth <atoth@atoth.sote.hu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>		/* strtoul(), malloc(), free(), exit() */
#include <string.h>		/* memset(), strcmp(), strncmp() */
#include <stdio.h>		/* printf(), fprintf() */
#include <sysexits.h>           /* EX_* */
#include <syslog.h>             /* openlog(), syslog(), closelog() */
#include <sys/types.h>		/* getaddrinfo(), freeaddrinfo(), fork(), open() */
#include <netdb.h>              /* getaddrinfo(), freeaddrinfo() */
#include <netinet/in.h>         /* struct sockaddr_in */
#include <ctype.h>              /* isspace() */
#include <unistd.h>		/* fork(), setsid(), close(), gethostname(), getdomainname() */
#include <sys/stat.h>		/* open() */
#include <fcntl.h>		/* open() */
#include <errno.h>		/* errno */
#include <limits.h>             /* UINT_MAX */

#include <libmilter/mfapi.h>

struct mlfiPriv
{
	char *mailing_list;
	char *subject;
	unsigned char listness;
	unsigned char subjectness;
	unsigned char clue;
};

char **list_array = NULL;

int listno = 0;
/* The following are all options set on the command line */
unsigned char debug = 0;
char *sock_addr = NULL;
char *conf_file = NULL;
char *host_name = NULL;
char *domain_name = NULL;
char *fqdn = NULL;

/*
 *   Initialization for the message
 */
sfsistat mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv;

	priv = malloc(sizeof *priv);
	if (priv == NULL) {
		syslog(LOG_DEBUG, "Malloc for listname_buffer failed\n");
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	priv->mailing_list = NULL;
	priv->subject = NULL;
	priv->listness = 0;
	priv->subjectness = 1;
	priv->clue = 0;

	smfi_setpriv(ctx, priv);

	return SMFIS_CONTINUE;
}

/* 
 *   Check for list addresses and set the necessary fields upon match
 */
sfsistat mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *rcpt_addr = NULL;
/*	char *rcpt_host = NULL; */
	int i;
	size_t len;

	len = (strlen(smfi_getsymval(ctx, "{rcpt_addr}")) + 1);
	rcpt_addr = realloc(rcpt_addr, len * sizeof(char));
	if (rcpt_addr == NULL) {
		syslog(LOG_DEBUG, "Malloc for addr_rcpt failed\n");
		return SMFIS_TEMPFAIL;
	}
	snprintf(rcpt_addr, len, "%s", smfi_getsymval(ctx, "{rcpt_addr}"));

/*	len = (strlen(smfi_getsymval(ctx, "{rcpt_host}")) + 1);
	rcpt_host = realloc(rcpt_host, len * sizeof(char));
	if (rcpt_host == NULL) {
		syslog(LOG_DEBUG, "Malloc for addr_host failed\n");
		return SMFIS_TEMPFAIL;
	}
	snprintf(rcpt_host, len, "%s", smfi_getsymval(ctx, "{rcpt_host}")); */

	if (debug == 1)
		syslog(LOG_DEBUG, "List-Milter: rcpt_addr: %s, rcpt_addr: %s, priv->listness: %d, priv->subjectness: %d\n", smfi_getsymval(ctx, "{rcpt_addr}"), rcpt_addr, priv->listness, priv->subjectness);

/*	if (strcmp(rcpt_host, fqdn) == 0)
	{ */
		for (i = 0; i < listno; i++)
		{
			if (priv->clue == 0)
			{
				if (strcmp(rcpt_addr, list_array[i]) == 0) {
					priv->clue = 1;

/*					if (debug == 1) { */
					if (debug == 1)
						syslog(LOG_DEBUG, "List-Milter: list_array[%d]: %s, listness: %d\n", i, list_array[i], priv->listness);
/*						syslog(LOG_DEBUG, "List-Milter: rcpt_addr: %s, listname_buffer: %s\n", rcpt_addr, listname_buffer);
					} */
				}
			}
		}
/*	} */
	free(rcpt_addr);
/*	free(rcpt_host); */

	if (debug == 1)
		syslog(LOG_DEBUG, "rcpt_addr: %s\n", smfi_getsymval(ctx, "{rcpt_addr}"));

	return SMFIS_CONTINUE;
}

/*
 *   Catching list message
 */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *listname_buffer = NULL;
	int i, k;
	size_t len;

	if (debug == 1)
		syslog(LOG_DEBUG, "List-Milter: headerf: %s headerv: %s, listno: %d", headerf, headerv, listno);

	if (strcmp("To", headerf) == 0)
	{
		if (priv->clue == 1)
		{
			for (i = 0; i < listno; i++)
			{
				if (priv->listness == 0)
				{
					len = (strlen(list_array[i]) + strlen(fqdn) + 2);
					listname_buffer = (char*) realloc(listname_buffer, len * sizeof(char));
					if (listname_buffer == NULL)
					{
						syslog(LOG_DEBUG, "Malloc for listname_buffer failed\n");
						return SMFIS_TEMPFAIL;
					}
					snprintf(listname_buffer, len, "%s@%s", list_array[i], fqdn);

					if (debug == 1)
						syslog(LOG_DEBUG, "List-Milter: headerv: %s listname_buffer: %s", headerv, listname_buffer);

					if (strstr(headerv, listname_buffer) == NULL)
					{
						priv->mailing_list = NULL;
						priv->listness = 0;
					}
					else
					{
						priv->mailing_list = list_array[i];
						priv->listness = 1;
					}
				}
			}
			free(listname_buffer);
		}
	}

	if (strcmp("Subject", headerf) == 0)
	{
		if (priv->clue == 1)
		{
/*			len = (strlen(priv->mailing_list) + 10);
			listname_buffer = realloc(listname_buffer, len * sizeof(char));
			if (listname_buffer == NULL) {
				syslog(LOG_DEBUG, "Malloc for listname_buffer failed\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(listname_buffer, len, "[Atoth: %s]", priv->mailing_list); */

			if (strstr(headerv, "[Atoth: ") == NULL)
			{
				priv->subjectness = 0;
				if(priv->subject != NULL)
					free(priv->subject);
				priv->subject = strdup(headerv);
			}
			else
			{
				priv->subjectness = 1;
				if(priv->subject != NULL)
					free(priv->subject);
				priv->subject = NULL;
			}
		}
/*		free(listname_buffer); */
	}
	return SMFIS_CONTINUE;
}

/*
 *   Add header_content to the message, and add list headers if it is a list
 */
sfsistat mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);
	char *list_address_buf = NULL;
	char *list_header_buf = NULL;
	char *list_request_buf = NULL;
	size_t len;

	/* Add header_contents to the message */
	if (priv->listness == 1)
	{
		smfi_chgheader(ctx, "Precedence", 1, "list");

		len = (strlen(priv->mailing_list) + strlen(fqdn) + 2);
		list_address_buf = (char*) realloc(list_address_buf, len * sizeof(char));
		if (list_address_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_address_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_address_buf, len, "%s@%s", priv->mailing_list, fqdn);
		smfi_chgheader(ctx, "Reply-To", 1, list_address_buf);
		smfi_chgheader(ctx, "Reply-To-Mailing-List", 1, list_address_buf);
		smfi_chgheader(ctx, "X-Loop", 1, list_address_buf);
		smfi_chgheader(ctx, "X-BeenThere", 1, list_address_buf);

		len = (strlen(list_address_buf) + 11);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<%s-request@%s>", priv->mailing_list, fqdn);
		smfi_chgheader(ctx, "Return-Path", 1, list_header_buf);

		len = (strlen(list_address_buf) + 10);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<mailto:%s>", list_address_buf);
		smfi_chgheader(ctx, "List-Post", 1, list_header_buf);

		len = (strlen(list_address_buf) + 26);
		list_request_buf = (char*) realloc(list_request_buf, len * sizeof(char));
		if (list_request_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_request_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_request_buf, len, "<mailto:%s-request@%s?subject=", priv->mailing_list, fqdn);

		len = (strlen(list_request_buf) + 6);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%shelp>", list_request_buf);
		smfi_chgheader(ctx, "List-Help", 1, list_header_buf);

		len = (strlen(list_request_buf) + 11);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%ssubscribe>", list_request_buf);
		smfi_chgheader(ctx, "List-Subscribe", 1, list_header_buf);

		len = (strlen(list_request_buf) + 13);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%sunsubscribe>", list_request_buf);
		smfi_chgheader(ctx, "List-Unsubscribe", 1, list_header_buf);

		len = (strlen(list_address_buf) + 9);
		list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "<owner-%s@%s>", priv->mailing_list, fqdn);
		smfi_chgheader(ctx, "Errors-To", 1, list_header_buf);

		if (priv->subjectness == 0)
		{
			if (priv->subject == NULL)
			{
				syslog(LOG_DEBUG, "priv->subjectness = 0, but priv->subject is NULL!\n");
				return SMFIS_TEMPFAIL;
			}	
			len = (strlen(priv->mailing_list) + strlen(priv->subject) + 11);
			list_header_buf = (char*) realloc(list_header_buf, len * sizeof(char));
			if (list_header_buf == NULL)
			{
				syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
				return SMFIS_TEMPFAIL;
			}
			snprintf(list_header_buf, len, "[Atoth: %s] %s", priv->mailing_list, priv->subject);
			smfi_chgheader(ctx, "Subject", 1, list_header_buf);
			priv->subjectness = 1;
		}

		len = (strlen(priv->mailing_list) + 30);
		list_header_buf = realloc(list_header_buf, len * sizeof(char));
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, len, "%s mailing list headers handled", priv->mailing_list);
		smfi_chgheader(ctx, "X-List-Milter", 1, list_header_buf);
		syslog(LOG_DEBUG, "List-Milter: added header: %s", list_header_buf);

		syslog(LOG_DEBUG, "List-Milter - Prepended headers to message for mailing list: %s", priv->mailing_list);

		free(list_address_buf);
		free(list_header_buf);
		free(list_request_buf);
	}
	else
	{
		list_header_buf = realloc(list_header_buf, 14);
		if (list_header_buf == NULL)
		{
			syslog(LOG_DEBUG, "Malloc for list_header_buf failed\n");
			return SMFIS_TEMPFAIL;
		}
		snprintf(list_header_buf, 14, "non-list mail");
		smfi_chgheader(ctx, "X-List-Milter", 1, list_header_buf);
		syslog(LOG_DEBUG, "List-Milter: added header: %s", list_header_buf);

		free(list_header_buf);
	}
	return SMFIS_CONTINUE;
}

/*
 *   Free whatever I allocated
 */
sfsistat mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv;

	priv = smfi_getpriv(ctx);
/*	if (priv != NULL)
	{
		if (priv->listness == 1) {
			free(priv->mailing_list);
			free(priv->subject);
			priv->listness = 0;
			priv->subjectness = 1;
		}
		free(priv);
		smfi_setpriv(ctx, NULL);
	} */

	if (priv != NULL)
	{
		priv->mailing_list = NULL;
		if (priv->subject != NULL)
			free(priv->subject);
		priv->subject = NULL;
		priv->listness = 0;
		priv->subjectness = 1;
		priv->clue = 0;

		smfi_setpriv(ctx, NULL);
		free(priv);
	}

	if (debug == 1)
		syslog(LOG_DEBUG, "Connection closed.");

	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"list-milter",
	SMFI_VERSION,
	SMFIF_ADDHDRS | SMFIF_CHGHDRS,
	mlfi_connect,	/* xxfi_connect, */
	NULL,		/* xxfi_helo, */
	NULL,		/* xxfi_envfrom, */
	mlfi_envrcpt,	/* xxfi_envrcpt, */
	mlfi_header,	/* xxfi_header, */
	NULL,		/* xxfi_eoh, */
	NULL,		/* xxfi_body, */
	mlfi_eom,	/* xxfi_eom, */
	NULL,		/* xxfi_abort, */
	mlfi_close,	/* xxfi_close */
};

void print_usage(char* prog_name)
{
	printf("Usage:\n");
	printf("%s -p socket-Addr -c conf-file [-h] [-d]\n\n", prog_name);
	printf("  -p\tListen to the inet/unix socket defined as Socket-Addr\n");
	printf("  -c\tConf-File name to use\n");
	printf("  -h\tHelp, a.k.a. this message\n");
	printf("  -d\tDebug mode, currently only more verbose output to syslog\n");
	printf("\n\n");
}

void parse_cmdline(int argc, char **argv)
{
	int i;
	int help_needed = 0;
	int expecting_socket = 0;
	int expecting_conffile = 0;

	for (i = 1; i < argc; i++)
	{
		if (expecting_socket == 1)
		{
			sock_addr = argv[i];
			expecting_socket = 0;
			continue;
		}
		if (expecting_conffile == 1)
		{
			conf_file = argv[i];
			expecting_conffile = 0;
			continue;
		}
		if (strcmp(argv[i], "-d") == 0)
		{
			debug = 1;
			continue;
		}
		if (strcmp(argv[i], "-h") == 0)
		{
			help_needed = 1;
			break;
		}
		if (strcmp(argv[i], "-p") == 0)
		{
			expecting_socket = 1;
			continue;
		}
		if (strcmp(argv[i], "-c") == 0)
		{
			expecting_conffile = 1;
			continue;
		}
	}
	if ((help_needed == 1)
		|| (expecting_socket == 1)
		|| (expecting_conffile == 1)
/*		|| (conf_file == NULL) */
		|| (sock_addr == NULL))
	{
		print_usage(argv[0]);
		if (help_needed == 1)
			exit(EX_OK);
		else
			exit(EX_USAGE);
	}
}

void full_domainname()
{
	char *uts;
	int i, j, k;
	struct hostent *machine;
	size_t len = 1;

	host_name = (char *)malloc(len);
	while(gethostname(host_name, len) == -1 && errno == ENAMETOOLONG)
		host_name = (char *)realloc(host_name, ++len);
	if (debug == 1)
		printf("Hostname: %s\n", host_name);

	len = HOST_NAME_MAX;
	domain_name = (char *)malloc(len);
	while(getdomainname(domain_name, len) == -1 && errno == EINVAL)
		domain_name = (char *)realloc(domain_name, --len);
	if (debug == 1)
		printf("Domainname: %s\n", domain_name);

	machine = malloc(sizeof *machine);
	if (machine == NULL)
	{
		fprintf(stderr, "Malloc machine failed\n");
		exit;
	}
 	memset(machine, '\0', sizeof *machine);
	machine = gethostbyname(host_name);
	if (machine == NULL) {
		fprintf(stderr, "Gethostbyname failed\n");
		exit;
	}
	len = ((strlen(machine->h_name) + 1) * sizeof(char));
	if ((fqdn = (char*) malloc(len)) == NULL)
	{
		fprintf(stderr, "Malloc for fqdn failed\n");
		exit;
	}
	snprintf(fqdn, len, "%s", machine->h_name);
	if (debug == 1)
		printf("FQDN: %s, fqdn: %s\n\n", machine->h_name, fqdn);
}

void parse_confdefault()
{
	FILE *file_pointer;
	int i, k;
	char listname_buffer[HOST_NAME_MAX];

	file_pointer = fopen("/etc/mail/aliases", "r");
	if (file_pointer == NULL)
	{
		fprintf(stderr, "fopen(/etc/mail/aliases) failed\n");
		exit;
	}

	for (i = 0, listno = 1, k = 0; k != EOF; i++, listno++)
	{
		switch (k = fscanf(file_pointer, "owner-%63[^#@:\n]\n", listname_buffer))
		{
			case 1:
				list_array = realloc(list_array, listno * sizeof *list_array);
				if (list_array == NULL)
					fprintf(stderr, "Realloc for list_array failed\n");
				list_array[i] = malloc((strlen(listname_buffer) + 1) * sizeof(char));
				if (list_array[i] == NULL)
					fprintf(stderr, "Malloc for list_array[%d] failed\n", i);
				snprintf(list_array[i], (strlen(listname_buffer) + 1), "%s", listname_buffer);

				if (debug == 1)
					printf("list_array[%d]: %s\n", listno, list_array[i]);
				break;
			case 0:
				fscanf(file_pointer, "%*[^\n]\n");
				i--;
				listno--;
				break;
			case EOF:
				printf("k = EOF: %d\n", k);
				i--;
				listno--;
				break;
		}
	}
	fclose(file_pointer);
	listno = i;
	if (debug == 1)
		printf("listno: %d\n", listno);
}

void parse_conffile()
{
	FILE *file_pointer;
	int i, k;
	char listname_buffer[HOST_NAME_MAX];
	size_t len;

	file_pointer = fopen(conf_file, "r");
	if (file_pointer == NULL)
	{
		fprintf(stderr, "fopen(%s) failed\n", conf_file);
		exit;
	}

/*	i = 0;
	k = 0;
	while (k != EOF) */
	for (i = 0, listno = 1, k = 0; k != EOF; i++, listno++)
	{
		switch (k = fscanf(file_pointer, "%63[^#@\n]\n", listname_buffer))
		{
			case 1:
				list_array = realloc(list_array, listno * sizeof *list_array);
				if (list_array == NULL)
					fprintf(stderr, "Realloc for list_array failed\n");
				len = (strlen(listname_buffer) + 1);
				list_array[i] = malloc(len * sizeof(char));
				if (list_array[i] == NULL)
					fprintf(stderr, "Malloc for list_array[%d] failed\n", i);
				snprintf(list_array[i], len, "%s", listname_buffer);
/*				i++; */

				if (debug == 1)
					printf("list_array[%d]: %s, listno: %d\n", i, list_array[i], listno);
/*					printf("list_array[%d]: %s\n", i, list_array[i]); */
				break;
			case 0:
				fscanf(file_pointer, "%*[^\n]\n");
				i--;
				listno--;
				break;
			case EOF:
				if (debug == 1)
					printf("k = EOF: %d\n", k);
				i--;
				listno--;
				break;
		}
	}
	fclose(file_pointer);
	listno = i;
	if (debug == 1)
		printf("listno: %d\n", listno);
}

/*
 *   Standard way to daemonize, fork twice to make sure the final child is not a
 *   session leader. Then you have to close all open file descriptors to detach
 *   from the console, also change directory somewhere safe and reset umask.
 *   TODO: Error checking
 */
int daemonize()
{
	int r;

	r = fork();
	if (r < 0)		/* failed */
		return -1;
	else if (r != 0)	/* parent */
		exit(EX_OK);
	else			/* child */
	{
		setsid();
		r = fork();
		if (r < 0)		/* failed */
			return -1;
		else if (r != 0)	/* parent */
			exit(EX_OK);
		else			/* child */
		{
			close(0); close(1); close(2);	/* close old fds */
			open("/dev/null", O_RDWR);	/* new fd 0 */
			dup(0); dup(0);			/* and fds 1 and 2 */
			umask(0);			/* clear umask */
			chdir("/var/tmp");		/* change to /var/tmp */
			return 0;
		}
	}
}

/*
 *   A milter's main() has to do only 3 things:
 *   1) Set the listening socket with smfi_setcon()
 *   2) Register the callbacks table with smfi_register()
 *   3) Pass control to smfi_main
 *
 *   Simple, isn't it?
 */
int main(int argc, char **argv) 
{
	int rc;

	parse_cmdline(argc, argv);

	if (conf_file == NULL)
		parse_confdefault();
	else
		parse_conffile();

	full_domainname();

	openlog(argv[0], LOG_PID | LOG_NDELAY, LOG_MAIL);

	smfi_setconn(sock_addr);

	rc = smfi_register(smfilter);
	if (rc == MI_FAILURE)
	{
		fprintf(stderr,"smfi_register() failed\n");
		exit(EX_UNAVAILABLE);
	}

	if (daemonize() != 0)
	{
		fprintf(stderr, "daemonization failed\n");
		exit(EX_UNAVAILABLE);
	}

	rc = smfi_main();
/*	smfi_stop(); */
	closelog();
	return rc;
}
